--{-# OPTIONS -fglasgow-exts #-}

--monadic interpreter
module CircInterpreter where
import Id
import ErrorT
import EnvT
import AuxTypes
import Control.Monad
import Data.Maybe

import Language
import TypeChecker
import Matrix
import Data.Complex
import Data.Array

import Interpreter
import QCirc
import QFqc

circCon :: CheckedExpr -> Fqc
circCon (CE n (AVar x) m t) = varFqc n
circCon (CE n (ASuper c1 c2) m t) = supFqc n c1 c2
circCon (CE n (ALet x ce1 ce2) m t) = nseqFqc (circCon ce1) (circCon ce2)
circCon (CE n (AProd ce1 ce2) m t) = nseqFqc (circCon ce1) (circCon ce2)
circCon (CE n (ALetP x1 x2 ce1 ce2) m t) = nseqFqc (circCon ce1) (circCon ce2)
circCon (CE n (AIf ce ce1 ce2) m t) = let k = case typ ce of
                                                Qbit n -> n
                                                _ -> error "circCon: invalid type for if condition"
                                      in ifFqc n (circCon ce) (circCon ce1) (circCon ce2) k
circCon (CE n (AIfM ce ce1 ce2) m t) = let k = case typ ce of
                                                Qbit n -> n
                                                _ -> error "circCon: invalid type for if condition"
                                       in ifmFqc n (circCon ce) (circCon ce1) (circCon ce2) k
circCon (CE n (ATrs ce x x' c) m t) = let b = qbits t
                                          siz = length b
                                          m = calc x x' c t
                                      in transFqc n (circCon ce) m b

--Auxiliary
bool2Cval :: [Bool] -> CValue
bool2Cval [b] = Bit b
bool2Cval (b:bs) = Pair (Bit b) (bool2Cval bs)
