{-# LANGUAGE FlexibleInstances #-}

--monadic interpreter
module Interpreter where
import Id
import ErrorT
import EnvT
import AuxTypes
import Control.Monad
import Data.Maybe

import Language
import TypeChecker
import Matrix
import Data.Complex
import Data.Array

ev				:: CheckedExpr -> (Matrix (Complex Double) -> Matrix (Complex Double))
ev (CE n (ALet x ce1 ce2) m t)	= ev ce2 . ev ce1
ev (CE n (AProd ce1 ce2) m t)	= ev ce2 . ev ce1
ev (CE n (ALetP x1 x2 ce1 ce2) m t)	= ev ce2 . ev ce1
ev (CE n (AIf ce ce1 ce2) m t) = \aA -> let m1 = exprbits ce1
				            m2 = exprbits ce2
				            k = case typ ce of
				                  Qbit n -> n
				                  _ -> error "ev: invalid type for if condition"
				            bB = ev ce aA
				            tT1 = ev0 ce1
				            tT2 = ev0 ce2
				            tT1' = except k tT1 |@| i (max m1 m2 - m1)
				            tT2' = except k tT2 |@| i (max m1 m2 - m2)
				            tTc = cond k tT1' tT2'
					 in tTc |*| (bB |@| d (max m1 m2)) |*| star tTc
ev (CE n (AIfM ce ce1 ce2) m t)	= \aA -> let m1 = exprbits ce1
				             m2 = exprbits ce2
				             bB = ev ce aA
				             k = case typ ce of
				                   Qbit n -> n
				                   _ -> error "ev: invalid type for if condition"
				             (bBt, bBf) = measure k bB
				             bB1 = ev ce1 bBt
				             bB2 = ev ce2 bBf
				         in bB1 |@| d (max m1 m2 - m1) |+| bB2 |@| d (max m1 m2 - m2)
ev (CE n (ATrs ce x x' c) m t)	= \aA -> let bB = ev ce aA
				  	     cC = calc x x' c t
                                             tT = expand (n+m) (qbits t) cC
				         in  tT |*| bB |*| star tT
ev ce				= \aA -> let m = exprbits ce
                                             tT = ev0 ce
				         in  tT |*| (aA |@| d m) |*| star tT

ev0				:: CheckedExpr -> Matrix (Complex Double)
ev0 (CE n (AVar x) m t)		= i n
ev0 (CE n (ASuper l l') m t)	= i n |@| Mqua (Msca l) (Msca l') (Msca l') (Msca (- l))
ev0 (CE n (ALet x1 ce1 ce2) m t)
				= tT2 |*| (tT1 |@| i m2)
				  where m2 = exprbits ce2
				        tT1 = ev0 ce1
					tT2 = ev0 ce2
ev0 (CE n (AProd ce1 ce2) m t)	= tT2 |*| (tT1 |@| i m2)
				  where m2 = exprbits ce2
				        tT1 = ev0 ce1
					tT2 = ev0 ce2
ev0 (CE n (ALetP x1 x2 ce1 ce2) m t)	= tT2 |*| (tT1 |@| i m2)
					  where m2 = exprbits ce2
					        tT1 = ev0 ce1
						tT2 = ev0 ce2
ev0 (CE n (AIf ce ce1 ce2) m t) = tTc |*| (tT |@| i (max m1 m2))
				  where m1 = exprbits ce1
				        m2 = exprbits ce2
				        k = case typ ce of
				              Qbit n -> n
				              _ -> error "ev: invalid type for if condition"
				        tT = ev0 ce
				        tT1 = ev0 ce1
				        tT2 = ev0 ce2
				        tT1' = except k tT1 |@| i (max m1 m2 - m1)
				        tT2' = except k tT2 |@| i (max m1 m2 - m2)
				        tTc = cond k tT1' tT2'
ev0 (CE n (ATrs ce x x' c) m t)	= expand (n+m) (qbits t) cC |*| tT
				  where tT = ev0 ce
				  	cC = calc x x' c t

calc	:: I -> I -> CExpr -> QType -> Matrix (Complex Double)
calc x x' c t	= toMatrix (0,0) k a
		  where a = array ((0, 0), (2^k-1, 2^k-1))
				[((j,i), evc c (x, val t i) (x', val t j)) | (i,j) <- range ((0, 0), (2^k-1, 2^k-1))]
		        k = length (qbits t)

toMatrix		:: (Int, Int) -> Int -> Array (Int, Int) a -> Matrix a
toMatrix (x, y) 0 a	= Msca (a!(x,y))
toMatrix (x, y) k a	= Mqua (toMatrix (x, y) (k-1) a)
			       (toMatrix (x, y + 2^(k-1)) (k-1) a)
			       (toMatrix (x + 2^(k-1), y) (k-1) a)
			       (toMatrix (x + 2^(k-1), y + 2^(k-1)) (k-1) a)

val			:: QType -> Int -> CValue
val (Qbit k) n	| n == 0	= Bit False
		| n == 1 	= Bit True
		| otherwise	= error ("val: bad number n=" ++ show n)
val (Tens qt1 qt2) n	= Pair (val qt1 (n `div` (2^k))) (val qt2 (n `mod` (2^k)))
			  where k = sizeofC (toClassic qt2)

sizeofC			:: CType -> Int
sizeofC CBit		= 1
sizeofC (CProd ct1 ct2)	= sizeofC ct1 + sizeofC ct2

data	CValue	= Bit Bool
		| Pair CValue CValue
		| Com (Complex Double)
		deriving (Eq, Show)

evc			:: CExpr -> (I, CValue) -> (I, CValue) -> Complex Double
evc c p1 p2		= let cc = run (do {e <- rdEnv; inEnv (extendEnv p1 (extendEnv p2 e)) (cev c)})
			  in case cc of
			       Ok (Com z) -> z
			       Ok (_) -> error "evc: returned something other than complex"
			       Error s -> error s

cev			:: CExpr -> Compute CValue
cev (VarC x)		= do e <- rdEnv
			     return (fromJust (fromE e x))
cev (Con l)		= return (Com l)
cev (LetC x c1 c2)	= do e <- rdEnv
			     cv1 <- cev c1
			     inEnv (extendEnv (x, cv1) e) (cev c2)
cev (ProdC c1 c2)	= do cv1 <- cev c1
			     cv2 <- cev c2
			     return (Pair cv1 cv2)
cev (LetPC x1 x2 c1 c2)	= do e <- rdEnv
			     p <- cev c1
			     case p of
			     	Pair cv1 cv2 -> inEnv (extendEnv (x2, cv2) (extendEnv (x1, cv1) e)) (cev c2)
cev (IntC c)		= do cv <- cev c
                             let intToComplex n = fromInteger (toInteger n) :+ 0
			     return (Com (intToComplex (code cv)))
cev (Neg c)		= do {cv <- cev c; neg cv}
cev (Eq c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; eq cv1 cv2}
cev (Add c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; plus cv1 cv2}
cev (Sub c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; minus cv1 cv2}
cev (Mul c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; mul cv1 cv2}
cev (Div c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; div' cv1 cv2}
cev (Pow c1 c2)		= do {cv1 <- cev c1; cv2 <- cev c2; power cv1 cv2}
cev (IfC c c1 c2)	= do c <- cev c
			     case c of
			       Bit True -> cev c1
			       Bit False -> cev c2


code			:: CValue -> Int
code (Bit True)		= 1
code (Bit False)	= 0
code (Pair cv1 cv2)	= 2^k * code cv1 + code cv2
			  where sizeof (Bit _) = 1
                                sizeof (Pair cv1 cv2) = sizeof cv1 + sizeof cv2
                                k = sizeof cv2

plus			:: CValue -> CValue -> Compute CValue
plus (Com z1) (Com z2)	= return (Com (z1+z2))

minus			:: CValue -> CValue -> Compute CValue
minus (Com z1) (Com z2)	= return (Com (z1-z2))

mul			:: CValue -> CValue -> Compute CValue
mul (Com z1) (Com z2)	= return (Com (z1*z2))

div'			:: CValue -> CValue -> Compute CValue
div' (Com z1) (Com z2)	= return (Com (z1/z2))

power			:: CValue -> CValue -> Compute CValue
power (Com z1) (Com z2)	= return (Com (pow z1 z2))

neg			:: CValue -> Compute CValue
neg (Com z)		= return (Com (- z))

eq			:: CValue -> CValue -> Compute CValue
eq cv1 cv2		= return (Bit (cv1 == cv2))



type	Compute a	= EnvT Env (ErrorT Id) a

newtype	Env		= MkE {fromE :: I -> Maybe CValue}

instance EnvType Env where
 initialEnv		= MkE (\_ -> Nothing)

extendEnv		:: (I, CValue) -> Env -> Env
extendEnv (i, v) e	= MkE (\i' -> if i==i' then Just v else fromE e i')

------------------------------------------------------------------
instance ErrMonad m => ErrMonad (EnvT Env m) where
	err		= lift . err
--------------------------------------------------------------
