module Language where

import Data.Complex

type I			= String

data Expr		= Var I
			| Super (Complex Double) (Complex Double)
			| Let I Expr Expr
			| Prod Expr Expr
			| LetP I I Expr Expr
			| If Expr Expr Expr
			| IfM Expr Expr Expr
			| Trs Expr I I CExpr
                        | App I [Expr]                  -- for macros!
			deriving (Eq, Read, Show)

data CExpr              = VarC I
			| Con (Complex Double)
			| LetC I CExpr CExpr
			| ProdC CExpr CExpr
			| LetPC I I CExpr CExpr
			| IntC CExpr
                        | Neg CExpr
                        | Eq CExpr CExpr
			| Add CExpr CExpr
			| Sub CExpr CExpr
			| Mul CExpr CExpr
			| Div CExpr CExpr
			| Pow CExpr CExpr
			| IfC CExpr CExpr CExpr
                        | AppC I [CExpr]                -- for macros!
			deriving (Eq, Read, Show)

data QType		= Qbit Int
			| Tens QType QType
			deriving (Eq, Read, Show)

data CType		= CBit
			| CProd CType CType
			| CComplex
			deriving (Eq, Read, Show)

type Type		= (QType, Int)

data AnnExpr a		= AVar I
			| ASuper (Complex Double) (Complex Double)
			| ALet I a a
			| AProd a a
			| ALetP I I a a
			| AIf a a a
			| AIfM a a a
			| ATrs a I I CExpr
			deriving (Eq, Read, Show)

data CheckedExpr	= CE {envbits::Int, expr::(AnnExpr CheckedExpr),
                              exprbits::Int, typ::QType}
			deriving (Eq, Read, Show)

data Body = E Expr | C CExpr
  deriving Show

type Subst = [(I, ([I], Body))]

without :: [I] -> Subst -> Subst
without is = filter (\(i, _) -> not (i `elem` is))

subst0 :: Subst
subst0 = [ ("e", ([], C (Con (exp 1)))), ("pi", ([], C (Con pi))) , ("i", ([], C (Con (0 :+ 1)))) ]

subst :: Subst -> Expr -> Expr
subst sigma expr =
   case expr of
      Var i ->
         case lookup i sigma of
            Just ([], E e) -> subst sigma e
            _              -> expr
      Super _ _ ->
         expr
      Let i e1 e2 ->
         let sigma' = without [i] sigma
         in  Let i (subst sigma e1) (subst sigma' e2)
      Prod e1 e2 ->
         Prod (subst sigma e1) (subst sigma e2)
      LetP i j e1 e2 ->
         let sigma' = without [i, j] sigma
         in  LetP i j (subst sigma e1) (subst sigma' e2)
      If e e1 e2 ->
         If (subst sigma e) (subst sigma e1) (subst sigma e2)
      IfM e e1 e2 ->
         IfM (subst sigma e) (subst sigma e1) (subst sigma e2)
      Trs e i j c ->
         let sigma' = without [i, j] sigma in
         Trs (subst sigma e) i j (substC sigma' c)
      App i es ->
         case lookup i sigma of
            Just ([], E (Var i')) ->
               subst sigma (App i' es)
            Just (is, E e) ->
               let walk [] [] sigma = sigma
                   walk (i : is) (e : es) sigma =
                      (i, ([], E e)) : walk is es sigma
                   walk _ _ _ =
                       error ("invalid macro parameters for " ++ i ++
                              " in " ++ show expr)
                   sigma' = walk is es []
                   e' = subst sigma' e
               in  subst sigma e'
            Just _ -> error ("wrong use of definition for " ++ i ++
                             " in " ++ show expr)
            _ -> App i (map (subst sigma) es)

substC :: Subst -> CExpr -> CExpr
substC sigma cexpr =
   case cexpr of
      VarC i ->
         case lookup i sigma of
            Just ([], C c) -> substC sigma c
            _              -> cexpr
      Con _ ->
         cexpr
      LetC i c1 c2 ->
         let sigma' = without [i] sigma
         in  LetC i (substC sigma c1) (substC sigma' c2)
      ProdC c1 c2 ->
         ProdC (substC sigma c1) (substC sigma c2)
      LetPC i j c1 c2 ->
         let sigma' = without [i, j] sigma
         in  LetPC i j (substC sigma c1) (substC sigma' c2)
      IntC c ->
         IntC (substC sigma c)
      Neg c ->
         Neg (substC sigma c)
      Eq c1 c2 ->
         Eq (substC sigma c1) (substC sigma c2)
      Add c1 c2 ->
         Add (substC sigma c1) (substC sigma c2)
      Sub c1 c2 ->
         Sub (substC sigma c1) (substC sigma c2)
      Mul c1 c2 ->
         Mul (substC sigma c1) (substC sigma c2)
      Div c1 c2 ->
         Div (substC sigma c1) (substC sigma c2)
      Pow c1 c2 ->
         Pow (substC sigma c1) (substC sigma c2)
      IfC c c1 c2 ->
         IfC (substC sigma c) (substC sigma c1) (substC sigma c2)
      AppC i cs ->
         case lookup i sigma of
            Just ([], C (VarC i')) ->
               substC sigma (AppC i' cs)
            Just (is, C c) ->
               let walk [] [] sigma = sigma
                   walk (i : is) (c : cs) sigma =
                      (i, ([], C c)) : walk is cs sigma
                   walk _ _ _ =
                       error ("invalid macro parameters for " ++ i ++
                              " in " ++ show cexpr)
                   sigma' = walk is cs []
                   c' = substC sigma' c
               in  substC sigma c'
            Just _ -> error ("wrong use of definition for " ++ i ++
                             " in " ++ show cexpr)
            _ -> AppC i (map (substC sigma) cs)
