-- This is a Happy file, not a yacc/bison file!

{
module Parser (parse, lexer) where
import Data.Complex
import Data.Char
import Language
import Matrix (pow)
}

%name parse
%tokentype { Token }

%token
   var      { T_var $$ }
   float    { T_float $$ }
   
   'qtrue'  { T_qtrue }
   'qfalse' { T_qfalse }
   'let'    { T_let }
   'in'     { T_in }
   'if'     { T_if }
   'ifm'    { T_ifm }
   'then'   { T_then }
   'else'   { T_else }
   'int'    { T_int }
   'sqrt'   { T_sqrt }
   'def'    { T_def }
   'defc'   { T_defc }

   '='      { T_eq }
   '('      { T_lparen }
   ')'      { T_rparen }
   '{'      { T_lcurly }
   '}'      { T_rcurly }
   ','      { T_comma }
   '|'      { T_bar }
   '>'      { T_gt }
   '->'     { T_arrow }
   '.'      { T_dot }
   '+'      { T_add }
   '-'      { T_sub }
   '*'      { T_mul }
   '/'      { T_div }
   '^'      { T_pow }
   ';'      { T_semi }

%nonassoc  'if' 'ifm' 'then' 'else'
%right 'in'
%nonassoc '='
%left  '+' '-'
%left  '*' '/'
%right '^'
%left SIGN
%left  'int'

%%

Prog : Decl Expr                                    { subst $1 $2 }

Decl : {- empty -}                                  { subst0 }
     | 'def' var Params '=' Expr ';' Decl           { ($2, ($3, E $5)) : $7 }
     | 'defc' var Params '=' CExpr ';' Decl         { ($2, ($3, C $5)) : $7 }

Params : {- empty -}                                   { [] }
       | var Params                                    { $1 : $2 }

Expr : Expr0                                           { $1 }
     | var Expr0s                                      { App $1 $2 }
     | 'let' var '=' Expr 'in' Expr                    { Let $2 $4 $6 }
     | 'let' '(' var ',' var ')' '=' Expr 'in' Expr    { LetP $3 $5 $8 $10 }
     | 'if' Expr 'then' Expr 'else' Expr               { If $2 $4 $6 }
     | 'ifm' Expr 'then' Expr 'else' Expr              { IfM $2 $4 $6 }
     | '|' Expr '>' '->' var ',' var '.' CExpr         { Trs $2 $5 $7 $9 }

Expr0 : var                                            { Var $1 }
      | '{' Lambda 'qfalse' Opt Lambda 'qtrue' '}'     { Super $2 ($4 $5) }
      | '(' Expr ',' Expr ')'                          { Prod $2 $4 }
      | '(' Expr ')'                                   { $2 }
-- syntactic sugar
      | 'qfalse'                                       { Super 1.0 0.0 }
      | 'qtrue'                                        { Super 0.0 1.0 }

Expr0s : Expr0                                         { [$1] }
       | Expr0 Expr0s                                  { $1 : $2 }

CExpr : var                                            { VarC $1 }
     | Lambda0                                         { Con ($1) }
     | 'let' var '=' CExpr 'in' CExpr                  { LetC $2 $4 $6 }
     | '(' CExpr ',' CExpr ')'                         { ProdC $2 $4 }
     | 'let' '(' var ',' var ')' '=' CExpr 'in' CExpr  { LetPC $3 $5 $8 $10 }
     | 'int' CExpr                                     { IntC $2 }
     | '+' CExpr %prec SIGN                            { $2 }
     | '-' CExpr %prec SIGN                            { Neg $2 }
     | CExpr '=' CExpr                                 { Eq $1 $3 }
     | CExpr '+' CExpr                                 { Add $1 $3 }
     | CExpr '-' CExpr                                 { Sub $1 $3 }
     | CExpr '*' CExpr                                 { Mul $1 $3 }
     | CExpr '/' CExpr                                 { Div $1 $3 }
     | CExpr '^' CExpr                                 { Pow $1 $3 }
     | 'if' CExpr 'then' CExpr 'else' CExpr            { IfC $2 $4 $6 }
     | '(' CExpr ')'                                   { $2 }

Lambda0 : float                                        { $1 :+ 0.0 }
        | 'sqrt' '(' Lambda1 ')'                       { sqrt $3 }

Lambda : Lambda0                                       { $1 }
       | '(' Lambda1 ')'                               { $2 }

Lambda1 : Lambda                                       { $1 }
        | '+' Lambda1 %prec SIGN                       { $2 }
        | '-' Lambda1 %prec SIGN                       { - $2 }
        | Lambda1 '+' Lambda1                          { $1 + $3 }
        | Lambda1 '-' Lambda1                          { $1 - $3 }
        | Lambda1 '*' Lambda1                          { $1 * $3 }
        | Lambda1 '/' Lambda1                          { $1 / $3 }
        | Lambda1 '^' Lambda1                          { pow $1 $3 }

Opt : '+'  { \c -> c }
    | '-'  { \c -> -c }

{

happyError :: [Token] -> a
happyError _ = error "Parse error"

data Token = T_EOF    -- redundant, but I hate warnings!
           | T_var I
           | T_float Double
           | T_let
           | T_in
           | T_if
           | T_ifm
           | T_then
           | T_else
           | T_int
           | T_sqrt
           | T_lparen
           | T_rparen
           | T_lcurly
           | T_rcurly
           | T_comma
           | T_bar
           | T_eq
           | T_gt
           | T_arrow
           | T_dot
           | T_add
           | T_sub
           | T_mul
           | T_div
           | T_pow
           | T_qtrue
           | T_qfalse
           | T_def
           | T_defc
           | T_semi
  deriving Show

lexer :: String -> [Token]
lexer [] = []
lexer ('#':cs) =
    case span (\c -> c /= '\n') cs of
       (_, '\n':cs) -> lexer cs
       (_, "") -> []
lexer (c:cs) | isSpace c = lexer cs
             | isAlpha c = lexVar (c:cs)
             | isDigit c = lexNum (c:cs)
lexer ('-':'>':cs) = T_arrow : lexer cs
lexer ('(':cs) = T_lparen : lexer cs
lexer (')':cs) = T_rparen : lexer cs
lexer ('{':cs) = T_lcurly : lexer cs
lexer ('}':cs) = T_rcurly : lexer cs
lexer ('=':cs) = T_eq : lexer cs
lexer ('+':cs) = T_add : lexer cs
lexer ('-':cs) = T_sub : lexer cs
lexer ('*':cs) = T_mul : lexer cs
lexer ('/':cs) = T_div : lexer cs
lexer ('^':cs) = T_pow : lexer cs
lexer (',':cs) = T_comma : lexer cs
lexer ('.':cs) = T_dot : lexer cs
lexer ('|':cs) = T_bar : lexer cs
lexer ('>':cs) = T_gt : lexer cs
lexer (';':cs) = T_semi : lexer cs
lexer (c:cs) = error ("invalid token " ++ show c)

lexNum cs = reads cs >>= \(f, cs') -> T_float f : lexer cs'

lexVar cs =
   let inAlpha c = isAlpha c || isDigit c || c == '\''
   in  case span inAlpha cs of
          ("let", rest)    -> T_let : lexer rest
          ("in", rest)     -> T_in : lexer rest
          ("if", rest)     -> T_if : lexer rest
          ("ifm", rest)    -> T_ifm : lexer rest
          ("then", rest)   -> T_then : lexer rest
          ("else", rest)   -> T_else : lexer rest
          ("int", rest)    -> T_int : lexer rest
          ("sqrt", rest)   -> T_sqrt : lexer rest
          ("qtrue", rest)  -> T_qtrue : lexer rest
          ("qfalse", rest) -> T_qfalse : lexer rest
          ("def", rest)    -> T_def : lexer rest
          ("defc", rest)   -> T_defc : lexer rest
          (var, rest)      -> T_var var : lexer rest

}
