{-# LANGUAGE ScopedTypeVariables, TypeSynonymInstances, FlexibleInstances #-}

module TypeChecker(typecheck,typecheck0, toClassic, qbits) where
import Id
import ErrorT
import EnvT
import AuxTypes
import Control.Monad

import Language

import Data.Complex
import Data.List

type TypeCheck a	= EnvT StaticEnv (ErrorT Id) a

type Env		= [(I, QType)]
type StaticEnv		= (Env, Int)

instance EnvType StaticEnv where
	initialEnv	= ([], 0)

instance ErrMonad m => ErrMonad (EnvT StaticEnv m) where
	err		= lift .err

-- How do typecheck and typechek0 intertwine ? 
typecheck		:: Expr -> TypeCheck CheckedExpr
typecheck (Let x e1 e2)	= do (e, n) <- rdEnv
			     ce1 <- typecheck e1
			     let t1 = typ ce1
			         m1 = exprbits ce1
			     ce2 <- inEnv (extendEnv e (x, t1), n+m1) 
				  		(typecheck e2)
			     let t = typ ce2
			         m2 = exprbits ce2
			     return (CE n (ALet x ce1 ce2) (m1 + m2) t)
typecheck (Prod e1 e2)	= do ((e, n) :: StaticEnv) <- rdEnv
			     ce1 <- typecheck e1
			     let t1 = typ ce1
			         m1 = exprbits ce1
			     ce2 <- inEnv (e, n+m1) (typecheck e2)
			     let t2 = typ ce2
			         m2 = exprbits ce2
			     return (CE n (AProd ce1 ce2) (m1+m2) (Tens t1 t2))
typecheck (LetP x1 x2 e1 e2)	= 
			  do (e, n) <- rdEnv
			     ce1 <- typecheck e1
			     let tt = typ ce1
			         m1 = exprbits ce1
			     case tt of
			        Tens t1 t2 -> do ce2 <- inEnv (extendEnv (extendEnv e (x1, t1)) (x2, t2), n+m1) (typecheck e2)
			                         let t = typ ce2
						     m2 = exprbits ce2
			                         return (CE n (ALetP x1 x2 ce1 ce2) (m1+m2) t) 
			        _ -> err ("invalid expr in let prod")
typecheck (If e e1 e2)	= do (env, n) <- rdEnv
			     ce <- typecheck e
			     let tt = typ ce
			         m = exprbits ce
			     case tt of
			        Qbit k -> do ce1 <- inEnv (restrict env k, n+m) (typecheck e1)
				             let t1 = typ ce1
					         m1 = exprbits ce1
			                     ce2 <- inEnv (restrict env k, n+m) (typecheck e2)
				             let t2 = typ ce2
					         m2 = exprbits ce2
			                     if t1==t2
			                        then return (CE n (AIf ce ce1 ce2) (m + max m1 m2) t2)
			                        else err ("invalid expr in if")
			        _ -> err ("invalid cond expr in if")
typecheck (IfM e e1 e2)	= do ((env, n) :: StaticEnv) <- rdEnv
			     ce <- typecheck e
			     let tt = typ ce
			         m = exprbits ce
			     case tt of
			        Qbit k -> do ce1 <- inEnv (env, n+m) (typecheck e1)
				             let t1 = typ ce1
					         m1 = exprbits ce1
			                     ce2 <- inEnv (env, n+m) (typecheck e2)
				             let t2 = typ ce2
					         m2 = exprbits ce2
			                     if t1==t2
			                        then return (CE n (AIfM ce ce1 ce2) (m + max m1 m2) t2)
			                        else err ("invalid expr in if")
			        _ -> err ("invalid cond expr in if")
typecheck (Trs e x x' c)	=
			  do ((env, n) :: StaticEnv) <- rdEnv
			     ce <- typecheck e
			     let t = typ ce
			         m = exprbits ce
			     if (pure t)
			      then case ctypecheck [(x, toClassic t), (x', toClassic t)] c of
			              Just CComplex -> return (CE n (ATrs ce x x' c) m t)
			              _ -> err ("invalid c expr in trans")
			      else err ("not pure expression used in trans, of type: " ++ show t)
typecheck e		= typecheck0 e


typecheck0		:: Expr -> TypeCheck CheckedExpr
typecheck0 (Var x)	= do ((e,n) :: StaticEnv) <- rdEnv
			     case lookup x e of
				Just t -> return (CE n (AVar x) 0 t)
				Nothing -> err ("unbound variable: " ++ x)
typecheck0 (Super l l') = let m = magnitude l
                              m' = magnitude l'
                              mm = m*m + m'*m'
                              sqmm = sqrt mm :+ 0
                          in  if abs (mm - 1) < 0.01
                               then do ((_, n) :: StaticEnv) <- rdEnv
                                       return (CE n (ASuper (l/sqmm) (l'/sqmm)) 1 (Qbit n))
                               else err ("invalid complex constants in superposition: " ++ show l ++ ", " ++ show l')
typecheck0 (Let x e1 e2)	=
			  do (e, n) <- rdEnv
			     ce1 <- typecheck0 e1
			     let t1 = typ ce1
			         m1 = exprbits ce1
			     ce2 <- inEnv (extendEnv e (x, t1), n+m1) 
						(typecheck0 e2)
			     let t = typ ce2
			         m2 = exprbits ce2
			     return (CE n (ALet x ce1 ce2) (m1 + m2) t)
typecheck0 (Prod e1 e2)	= do ((e, n) :: StaticEnv) <- rdEnv
			     ce1 <- typecheck0 e1
			     let t1 = typ ce1
			         m1 = exprbits ce1
			     ce2 <- inEnv (e, n+m1) (typecheck0 e2)
			     let t2 = typ ce2
			         m2 = exprbits ce2
			     return (CE n (AProd ce1 ce2) (m1+m2) (Tens t1 t2))
typecheck0 (LetP x1 x2 e1 e2)	= 
			  do (e, n) <- rdEnv
			     ce1 <- typecheck0 e1
			     let tt = typ ce1
			         m1 = exprbits ce1
			     case tt of
			        Tens t1 t2 -> do ce2 <- inEnv (extendEnv (extendEnv e (x1, t1)) (x2, t2), n+m1) (typecheck0 e2)
			                         let t = typ ce2
						     m2 = exprbits ce2
			                         return (CE n (ALetP x1 x2 ce1 ce2) (m1+m2) t) 
			        _ -> err ("invalid expr in let prod")
typecheck0 (If e e1 e2)	= do (env, n) <- rdEnv
			     ce <- typecheck0 e
			     let tt = typ ce
			         m = exprbits ce
			     case tt of
			        Qbit k -> do ce1 <- inEnv (restrict env k, n+m) (typecheck0 e1)
			                     let t1 = typ ce1
					         m1 = exprbits ce1
			                     ce2 <- inEnv (restrict env k, n+m) (typecheck0 e2)
			                     let t2 = typ ce2
					         m2 = exprbits ce2
			                     if t1==t2
			                        then return (CE n (AIf ce ce1 ce2) (m + max m1 m2) t2)
			                        else err ("invalid expr in if")
			        _ -> err ("invalid cond expr in if")
typecheck0 (Trs e x x' c)	=
			  do ((env, n) :: StaticEnv) <- rdEnv
			     ce <- typecheck0 e
			     let t = typ ce
			         m = exprbits ce
			     if (pure t)
			      then case ctypecheck [(x, toClassic t), (x', toClassic t)] c of
			              Just CComplex -> return (CE n (ATrs ce x x' c) m t)
			              Nothing -> err ("invalid c expr in trans")
			      else err ("not pure expression used in trans")
typecheck0 (App i _)            = error ("application of " ++ i ++ " somehow survived!")

toClassic		:: QType -> CType
toClassic (Qbit _)	= CBit
toClassic (Tens t1 t2)	= CProd (toClassic t1) (toClassic t2)
   
qbits			:: QType -> [Int]
qbits (Qbit n)		= [n]
qbits (Tens t1 t2)	= (qbits t1) `union` (qbits t2)

pure			:: QType -> Bool
pure (Qbit n)		= True
pure (Tens t1 t2)	= pure t1 && pure t2 && (qbits t1 `intersect` qbits t2 == [])

restrict		:: Env -> Int -> Env
restrict ((i, t):es) k	= if elem k (qbits t) then restrict es k else (i, t):(restrict es k)

extendEnv		:: Env -> (I, QType) -> Env
extendEnv e (i, t)	= (i, t):e

ctypecheck			:: [(I, CType)] -> CExpr -> Maybe CType
ctypecheck d (VarC x)		= case lookup x d of
                                     Nothing -> error ("unbound variable " ++ x)
                                     x -> x
ctypecheck d (Con l)		= Just CComplex
ctypecheck d (LetC x c1 c2)	= case ctypecheck d c1 of
					Just f1 -> ctypecheck ((x, f1):d) c2
				        Nothing -> Nothing
ctypecheck d (ProdC c1 c2)	= case ctypecheck d c1 of
				    Just f1 -> case ctypecheck d c2 of
						    Just f2 -> Just (CProd f1 f2)
						    Nothing -> Nothing
				    Nothing -> Nothing
ctypecheck d (LetPC x1 x2 c1 c2)	=
				  case ctypecheck d c1 of
				       Just (CProd f1 f2) -> ctypecheck ((x1, f1):(x2, f2):d) c2 
				       _ -> Nothing
ctypecheck d (IntC c)		= case ctypecheck d c of
				    Just CBit -> Just CComplex
				    Just (CProd _ _) -> Just CComplex
				    _ -> Nothing
ctypecheck d (Neg c)	        = case ctypecheck d c of
				    Just CComplex -> Just CComplex
                                    _ -> Nothing
ctypecheck d (Eq c1 c2)	        = case (ctypecheck d c1, ctypecheck d c2) of
				    (Just t1, Just t2) -> if t1 == t2 then Just CBit else Nothing
                                    _ -> Nothing
ctypecheck d (Add c1 c2)	= case ctypecheck d c1 of
				    Just CComplex -> ctypecheck d c2
                                    _ -> Nothing
ctypecheck d (Sub c1 c2)	= case ctypecheck d c1 of
				    Just CComplex -> ctypecheck d c2
                                    _ -> Nothing
ctypecheck d (Mul c1 c2)	= case ctypecheck d c1 of
				    Just CComplex -> ctypecheck d c2
                                    _ -> Nothing
ctypecheck d (Div c1 c2)	= case ctypecheck d c1 of
				    Just CComplex -> ctypecheck d c2
                                    _ -> Nothing
ctypecheck d (Pow c1 c2)	= case ctypecheck d c1 of
				    Just CComplex -> ctypecheck d c2
                                    _ -> Nothing
ctypecheck d (IfC c c1 c2)	= case ctypecheck d c of
				    Just CBit -> let f1 = ctypecheck d c1
				                     f2 = ctypecheck d c2
				                 in case f1 of
						     Just t -> if f1==f2 then f1 else Nothing
						     Nothing -> Nothing
                                    _ -> Nothing
ctypecheck d (AppC i _)         = error ("application of " ++ i ++ " somehow survived!")